/*
 * $Id: $
 *
 * Module:  MOD_MGIRPC -- Apache Interface for MGIRPC framework
 * Created: 03-JUN-2009 22:56
 * Author:  tmr
 */

/***** TODO *******************************************************
 *  o graceful child cleanup on apachectl-stop or kill
 *    (is it possible? gtm_* won't allow us to handle signals!)
 ******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>
#include <stdbool.h>

#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <http_log.h>

#include <apr.h>
#include <apr_poll.h>
#include <apr_file_io.h>
#include "apr_arch_file_io_snippet.h"

#include "gtmxc_types.h"


#define MGIRPC$BUFLEN           1024

#define MGIRPC$SS_SUCCESS       true
#define MGIRPC$SS_ERROR         false

// XXX: load the configuration file instead of this!
#define MGIRPC$CITBL            MGIRPC$PIPROOT"/sapi/mod_mgirpc/calltab.ci"
#define MGIRPC$RTNS             ". "MGIRPC$PIPROOT"/sapi/mod_mgirpc "MGIRPC$PIPROOT"/zrtns/obj("MGIRPC$PIPROOT"/zrtns) "MGIRPC$PIPROOT"/srtns/obj("MGIRPC$PIPROOT"/srtns) "MGIRPC$PIPROOT"/crtns/obj("MGIRPC$PIPROOT"/crtns) "MGIRPC$PIPROOT"/mrtns/obj("MGIRPC$PIPROOT"/mrtns) "MGIRPC$PIPROOT"/irtns "MGIRPC$PIPROOT"/scartns/obj("MGIRPC$PIPROOT"/scartns) "MGIRPC$PIPROOT"/gtm_dist "MGIRPC$SRVROOT"rtns/framework/obj("MGIRPC$SRVROOT"rtns/framework) "MGIRPC$SRVROOT"rtns/application/obj("MGIRPC$SRVROOT"rtns/application)"
#define MGIRPC$GBLDIR           ""MGIRPC$PIPROOT"/gbls/mumps.gld"
//#define MGIRPC$EXTRTNS          ""

/* Should we override parent's environment? */
#define MGIRPC$FORCENV          1

const char * c_EnvTable [] = {
  "GTMCI",          MGIRPC$CITBL,
  "gtmgbldir",      MGIRPC$GBLDIR,
  "gtmroutines",    MGIRPC$RTNS,
  "gtmgbldir",      MGIRPC$GBLDIR,
  "GTMXC_extcall",  MGIRPC$PIPROOT"/toolset/extcall_V1.2/extcall.xc",
  "GTMXC_ipc",      MGIRPC$PIPROOT"/toolset/extcalls/ipc/ipc.xc",
  "GTMXC_htm",      MGIRPC$PIPROOT"/toolset/extcalls/htm/htm.xc",
  "GTMXC_CRYPTO",   MGIRPC$PIPROOT"/server/rtns/external/CRYPTO.xc",
  "SCAU_NUPS_TMP",  MGIRPC$PIPROOT"/gbls/tmp.dat",
  "SCAU_NUPS_UBG",  MGIRPC$PIPROOT"/gbls/ubg.dat",
  "SCAU_NUPS_TBLS", MGIRPC$PIPROOT"/gbls/tbls.dat",

  /*
  "GTMXC_alerts",   MGIRPC$EXTRTNS "/extcall_V1.2/alerts.xc",
  "GTMXC_extcall",  MGIRPC$EXTRTNS "/extcall_V1.2/extcall.xc",
  "GTMXC_mtm",      MGIRPC$EXTRTNS "/mtm_V2.4.5/mtm.xc",
  "GTMXC_mq",       MGIRPC$EXTRTNS "/mq_V1.0.1/mq.xc",
  */
  NULL
};

// GTMXC_CRYPTO=MGIRPC$SRVROOT/rtns/external/CRYPTO.xc

static unsigned int       m_Hits;

static int mgirpc_init (apr_pool_t *pool, apr_pool_t *log,
                      apr_pool_t *tmp, server_rec *srv);
static void mgirpc_child_init (apr_pool_t *p, server_rec *s);
static apr_status_t mgirpc_child_exit (void *data);
static int mgirpc_check_status (gtm_status_t status);
static void mgirpc_log_error (const char * msg, ...);
static void mgirpc_report_error (request_rec * r, const char * msg, ...);
static int mgirpc_handler (request_rec* r);
static void mgirpc_hooks (apr_pool_t* pool);
static int mgirpc_setenv (void * rec, const char * key, const char * value);
static char * mgirpc_strnfmt (request_rec * r, const char * msg, ...);
static void mgirpc_redir_stdio2client (request_rec * r);

static int mgirpc_init (apr_pool_t *pool, apr_pool_t *log,
                      apr_pool_t *tmp, server_rec *srv) {

  m_Hits = 0; char buf [512];
  ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, NULL,
                "MGIRPC module %06d initialized (home dir: %s)",
                getpid (), getcwd (buf, sizeof (buf)));

  int i;
  for (i = 0; c_EnvTable [i] != NULL; i += 2)
    setenv (c_EnvTable [i], c_EnvTable [i+1], MGIRPC$FORCENV);

  return OK;
}

static void mgirpc_child_init (apr_pool_t *p, server_rec *s) {

  ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, NULL,
                "MGIRPC child %06d initialization...", getpid ());

  gtm_status_t status;
  status = gtm_init();
  (void) mgirpc_check_status (status);

  status = gtm_ci ("init", "", "", "\r\n");
  (void) mgirpc_check_status (status);

  apr_pool_cleanup_register (p, s, mgirpc_child_exit, mgirpc_child_exit);
}

static apr_status_t mgirpc_child_exit (void *data) {

  ap_log_error (APLOG_MARK, APLOG_NOTICE, 0, NULL,
      "MGIRPC child %06d cleanup...", getpid ());

  gtm_status_t status;
  status = gtm_ci ("finalize");
  (void) mgirpc_check_status (status);

  status = gtm_exit ();
  (void) mgirpc_check_status (status);

  return OK;
}

int mgirpc_check_status (gtm_status_t status) {

  if (status != 0) {
    gtm_char_t msgbuf [MGIRPC$BUFLEN];
    gtm_zstatus (msgbuf, MGIRPC$BUFLEN);
    mgirpc_log_error ((const char *) msgbuf);
    //exit (status);
    return MGIRPC$SS_ERROR;
  }

  return MGIRPC$SS_SUCCESS;
}

void mgirpc_log_error (const char * msg, ...) {

  char emsg [512];
  va_list args; va_start (args, msg);

  vsnprintf (emsg, sizeof (emsg), msg, args);
  ap_log_error (APLOG_MARK, APLOG_WARNING, 0, NULL,
      "MGIRPC Error: PID=%d; Description: %s", getpid (), emsg);
}

void mgirpc_report_error (request_rec * r, const char * msg, ...) {

  char emsg [512];
  va_list args; va_start (args, msg);
  snprintf (emsg, sizeof (emsg),
    "MGIRPC Error: PID=%d\nDescription=%s\n", getpid (), msg);

  mgirpc_log_error (msg, args);

  ap_rputs (emsg, r);
}

static int mgirpc_setenv (void * rec, const char * key, const char * value) {

  request_rec* r = rec;
  gtm_status_t status = gtm_ci ("setparam", key, value);
  (void) mgirpc_check_status (status);

  return 1;
}

static char * mgirpc_req_read (request_rec* r) {

  int len, ret;
  char * res;
  char * p;

  /* Ask Apache to dechunk data if it is chunked. */
  ret = ap_setup_client_block (r, REQUEST_CHUNKED_DECHUNK);
  if (ret != OK) {
    mgirpc_log_error ("dechunk problems: %d", ret);
    return 0;
  }
  
  /* This function will send a 100 Continue response if the client is
   * waiting for that. If the client isn't going to send data, then this
   * function will return 0. */
  if (!ap_should_client_block (r)) len = 0;
  else len = r->remaining;

  res = p = apr_palloc (r->pool, len +1);
  res [len] = 0;
  
  while (len > 0) {
    /* Read data from the client. Returns 0 on EOF or error, the
     * number of bytes otherwise. */
    ret = ap_get_client_block (r, p, len);
    if (!ret) {
      /* mgirpc_log_error the problem */
      mgirpc_log_error ("Failed to read data from client.");

      /* log it and notify a client also... (???) */
      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r, "Failed to read data from client.");

      return 0;  /* HTTP_INTERNAL_SERVER_ERROR */
    }

    p += ret; len -= ret;
  }

  return res;
}

/* XXX: This does not work ;-( */
static void mgirpc_redir_stdio2client (request_rec * r) {

/*
  apr_pollfd_t pfd = r->connection->cs->pfd;
  apr_descriptor desc = pfd.desc;
  apr_file_t   * file = desc.f;
  int fd              = file->filedes;

  mgirpc_log_error ("Stdout override: before");
  if (dup2 (fd, STDOUT_FILENO) < 0) {
    mgirpc_log_error ("dup2 for stdout failed...");
    return;
  }

  mgirpc_log_error ("Stderr override: before");
  if (dup2 (fd, STDERR_FILENO) < 0) {
    mgirpc_log_error ("dup2 for stderr failed...");
    return;
  }
  mgirpc_log_error ("Stdout/stderr override done...");
*/
}

static bool mgirpc_ci_process (request_rec * r) {

  gtm_status_t status;
  gtm_char_t response [MGIRPC$BUFLEN];

  status = gtm_ci ("process", mgirpc_req_read (r));
  if (mgirpc_check_status (status) != MGIRPC$SS_SUCCESS) {
    mgirpc_report_error (r, "MGIRPC/process Internal error!");
    return false;
  }

  return true;
}

static char mgirpc_ch2uc (char x)
  { return (('a' <= x && x <= 'z')? x -'a' +'A' : x); }

static bool mgirpc_strcmp_nocs (char * a, char * b) {

  while (*a != '\0' || *b != '\0')
    if (mgirpc_ch2uc (*a++) != mgirpc_ch2uc (*b++)) return false;

  return ((*a == '\0' && *b == '\0')? true : false);
}

static bool mgirpc_process_hdrs (request_rec * r) {

  gtm_status_t status;
  gtm_char_t hdrkey [MGIRPC$BUFLEN]; //hdrkey [0] = '\0';
  gtm_char_t hdrval [MGIRPC$BUFLEN];
  while (true) {
    memset (hdrval, 0, sizeof (hdrkey));
    memset (hdrval, 0, sizeof (hdrval));

    status = gtm_ci ("hdrorder", hdrkey, hdrkey);
    if (mgirpc_check_status (status) != MGIRPC$SS_SUCCESS) {
      mgirpc_report_error (r, "MGIRPC/hdrorder Internal error!");
      continue;
    }

    if (strlen (hdrkey) < 1) break;

    status = gtm_ci ("gethdr", hdrval, hdrkey);
    if (mgirpc_check_status (status) != MGIRPC$SS_SUCCESS) {
      mgirpc_report_error (r, "MGIRPC/gethdr Internal error!");
      continue;
    }

    //mgirpc_log_error ("Header: %s => %s; (is Content-Type: %s)",
    //    hdrkey, hdrval, mgirpc_strcmp_nocs (hdrkey, "Content-Type")? "true" : "false");

    /* XXX: even if the hdrval is NOT "text/plain", we are still getting  *
     * "content-type: text/plain" or "Content-Type: text/plain" :-(       */
    if (mgirpc_strcmp_nocs (hdrval, "Content-Type"))
      ap_set_content_type (r, hdrval);
    else
      apr_table_set (r->headers_out, hdrkey, hdrval);
  }

  return true;
}

static bool mgirpc_process_errors (request_rec * r) {

  gtm_status_t status;
  gtm_char_t errdescr [MGIRPC$BUFLEN];
  status = gtm_ci ("error", errdescr);
  if (mgirpc_check_status (status) != MGIRPC$SS_SUCCESS) {
    mgirpc_report_error (r, "MGIRPC/error Internal error!");
    return false;
  }

  if (strlen (errdescr) > 0) {
    mgirpc_report_error (r, "MGIRPC Error Description: %s", errdescr);
    return false;
  }

  return true;
}

static bool mgirpc_generate_response (request_rec * r) {

  gtm_status_t status;
  gtm_char_t response [MGIRPC$BUFLEN];
  gtm_long_t i = 0;

  while (i >= 0) {
    status = gtm_ci ("resporder", &i, response);
    if (mgirpc_check_status (status) != MGIRPC$SS_SUCCESS) {
      mgirpc_report_error (r, "MGIRPC/resporder#%d Internal error!", i);
      return false;
    }

    ap_rputs (response, r);
  }

  return true;
}

static int mgirpc_handler (request_rec * r) {

  if (!r->handler || strcmp (r->handler, "mgirpc"))
    return DECLINED;

  m_Hits ++;
  time_t time_start, time_end, time_diff;

  mgirpc_redir_stdio2client (r);

  apr_table_do (mgirpc_setenv, r, r->subprocess_env, NULL);

  time_start = time (NULL);
  if (!mgirpc_ci_process (r)) return OK;
  time_end = time (NULL);

  apr_table_set (r->headers_out, "X-MGIRPC-Child-JobID",
                 mgirpc_strnfmt (r, "%08d", getpid ()));
  apr_table_set (r->headers_out, "X-MGIRPC-Child-Hits",
                 mgirpc_strnfmt (r, "%08d", m_Hits));
  apr_table_set (r->headers_out, "X-MGIRPC-Response-Duration",
                 mgirpc_strnfmt (r, "%lfs", difftime (time_end, time_start)));

  if (!mgirpc_process_errors (r)) return OK;

  mgirpc_process_hdrs (r);
  mgirpc_generate_response (r);

  return OK;
}

static char * mgirpc_strnfmt (request_rec * r, const char * msg, ...) {

  va_list args; va_start (args, msg);

/* *** XXX: Why does it allocate space only for a three chars?!
  char * res;
  res = apr_palloc (r->pool, 256);

  vsnprintf (res, sizeof (res), msg, args);

  return res;
*/
  return (char *) apr_pvsprintf (r->pool, msg, args);
}

static void mgirpc_hooks (apr_pool_t * pool) {

  ap_hook_post_config (mgirpc_init, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler (mgirpc_handler, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_child_init (mgirpc_child_init, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA mgirpc_module = {
  STANDARD20_MODULE_STUFF,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  mgirpc_hooks
};

// vim: fdm=syntax:fdn=1:tw=74:ts=2:syn=c
